#!/bin/bash
# This script uses the cryptographic techniques of salting and 
# key stretching to build brainwallet addresses that are much stronger
# against e.g. rainbow table attacks than the standard sha256(passphrase)
# kind of passwords that some sites use.  As long as you use a
# high-entropy salt, even fairly simple passphrases will result in
# strong hashes.
#
# The script uses BOTG functions from:
# https://bitcointalk.org/index.php?topic=23081.0
# and requires the openssl package to function.
#
# Don't forget to make scripts executable before trying to run them, e.g.:
# chmod +x scriptname.sh
#
# A good place to get salt strings is: https://www.grc.com/passwords.htm
# This script is in the public domain - and comes with no warranty

# If you like this script, a contribution to my bitcoin address
#
#        1H8eUP2SqBtY2MpgomXbmsb11VNoZSSxJ7
#
# would be appreciated!  This address was created using the BrainWallet script.
# Thanks. Dave

# Usage: 1) Put your salt in a one-line file called salt
#        2) Encrypt the salt file using the encryption script.
#           ./encrypt.sh salt
#           You only need to do this once!
#           Don't lose the salt.aes file - you will need it if you
#           want to regenerate your private key from your passphrase.
#           Type a password to encrypt the salt file twice, and remember it!
#        4) To create a bitcoin private key, run the brainwallet script.
#           ./brainwallet.sh <your passphrase goes here>
#           Note - if your passphrase includes spaces, it should be
#           surrounded by double quotes.  Ex:
#           ./brainwallet.sh "this is my long passphrase"
#        5) Enter the password from item 3) and out will pop your new
#           private key - one that you can generate again using
#           the passphrase you've saved in your brain. And one that 
#           is much stronger than the standard sha256(passphrase)
#           brainwallet implementations.  Enjoy!

base58=({1..9} {A..H} {J..N} {P..Z} {a..k} {m..z})

encodeBase58() {
    # 58 = 0x3A
    bc <<<"ibase=16; n=${1^^}; while(n>0) { n%3A ; n/=3A }" |
    tac |
    while read n
    do echo -n ${base58[n]}
    done
		echo -e "\n"
}

checksum() {
    xxd -p -r <<<"$1" |
    openssl dgst -sha256 -binary |
    openssl dgst -sha256 -binary |
    xxd -p -c 80 |
    head -c 8
}

hash256ToAddress() {	
	#printf "80$1$(checksum "80$1")"
	printf "%34s\n" "$(encodeBase58 "80$1$(checksum "80$1")")" |
	sed "y/ /1/"
}

privateKeyToWIF() {
    #hash256ToAddress $(openssl ec -text -noout -in data.pem | head -5 | tail -3 | fmt -120 | sed 's/[: ]//g')
		hash256ToAddress $hexkey
}

hash160ToAddress() {
    printf "%34s\n" "$(encodeBase58 "00$1$(checksum "00$1")")" |
    sed "y/ /1/"
}

publicKeyToAddress() {
    hash160ToAddress $(
    openssl ec -pubin -pubout -outform DER |
    tail -c 65 |
    hash160
    )
}

hash160() {
    openssl dgst -sha256 -binary |
    openssl dgst -rmd160 -binary |
    xxd -p -c 80
}

iter=256 # number of iterations in your key stretch
j=0

salt=`openssl aes-256-cbc -d -in salt.aes`

if [ -z $salt ]; then
  echo "Bad AES password"
	exit
else
  inthash=`echo "$1"`
  while [ $j -lt $iter ];
  do
		inthash=${inthash}_${salt}
    let j=j+1
    key=`echo "$inthash" | sha256sum`
    inthash=${key:0:64}
  done
  hexkey=${key:0:64}
fi

if [ ${hexkey:0:32} == "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF" ]; then
  echo "Pathological key - use a different passphrase"
  exit
fi

#echo "This is your private key in hexadecimal format"
#echo $hexkey       # This is your key in hexadecimal format
echo "This is your private key in base58 WIF format"
privateKeyToWIF
echo "Import this key into your wallet and the wallet will calculate your Bitcoin address to receive funds with"
